{-# LANGUAGE ScopedTypeVariables #-}
module Main where

import Prelude
   hiding (sum, map, foldr, (++), length, takeWhile, reverse, zip, mapM)
import Codec.Picture
import Data.Char
import System.IO
import Data.Vector.Unboxed
import qualified Prelude as P
import qualified Data.Vector as V
import qualified Data.List as Lis
import Data.String.Utils
import Data.IORef
import qualified Data.Text as Text
import Graphics.UI.Gtk as Gtk
import Graphics.UI.Gtk.Gdk.PixbufAnimation
import Graphics.UI.Gtk.Selectors.ColorButton
import qualified Data.Array.MArray as MArr
import qualified Data.Text as TXT
import qualified Control.Foldl as Fold
import Turtle
import Text.Show.Unicode

type Point = (Int, Int)

fruitSeisei = do
  fruitPrim <- iVecFromFileJP "fruit.txt"
  randrand <- iVecFromFile "randrand.txt"
  let
    fruit = V.cons "。" $ V.init $ V.init fruitPrim
    rand = V.map (\x -> read x :: Double) $ V.init randrand
    randInts = V.map (\d -> floor $ d * (fromIntegral $ V.length fruit)) rand
    fruitList = V.map (\n -> fruit V.! n) randInts
  oVecToFileJP fruitList "fruitListed.txt"
  putStrLn ""

mainGtk = do
    initGUI

    -- Data Initialization
    books <- initializeBooks
    (book, page) <- initBooketPage books
    booksRef <- newIORef books
    bookRef <- newIORef book
    pageRef <- newIORef page

    -- GUI Initialization
    window <- initMainWindow bookRef pageRef
    Gtk.on (wMain window) objectDestroy mainQuit
    colorpallete <- initColorPallete window bookRef pageRef booksRef
    movepallete <- initMovePallete window bookRef pageRef booksRef

    -- Balancing when events triggerred
    balancePage2 window colorpallete movepallete bookRef pageRef booksRef

    -- Events
    eTouch window colorpallete movepallete bookRef pageRef
    eRelease window colorpallete movepallete bookRef pageRef booksRef
    eMoveOtherPage window colorpallete movepallete bookRef pageRef booksRef (wButNext window) RightD
    eMoveOtherPage window colorpallete movepallete bookRef pageRef booksRef (wButPrev window) LeftD
    ePurgeCConfig window colorpallete movepallete bookRef pageRef booksRef (wButVanilla window)
    eEludeToken (cpButEludeToL colorpallete) (cpButShowToken colorpallete) bookRef pageRef  True
    eEludeToken (cpButEludeToR colorpallete) (cpButShowToken colorpallete) bookRef pageRef  False

    let
      eRegistCConfigFunc but color = do
        eRegistCConfig window colorpallete movepallete bookRef pageRef booksRef but color

    eRegistCConfigFunc (cpButRed colorpallete) "Red"
    eRegistCConfigFunc (cpButBlue colorpallete) "Blue"
    eRegistCConfigFunc (cpButGreen colorpallete) "Green"
    eRegistCConfigFunc (cpButPurple colorpallete) "Purple"
    eRegistCConfigFunc (cpButOrange colorpallete) "Orange"
    eRegistCConfigFunc (cpButPink colorpallete) "Pink"
    eRegistCConfigFunc (cpButBrown colorpallete) "Brown"
    eRegistCConfigFunc (cpButAqua colorpallete) "Aqua"

    -- Termination
    Gtk.on (wMain window) objectDestroy $ widgetDestroy (cpWindow colorpallete)
    Gtk.on (wMain window) objectDestroy $ widgetDestroy (mpWindow movepallete)
    widgetShowAll $ wMain window
    mainGUI

eEludeToken :: Gtk.Button -> Gtk.Button -> IORef Book -> IORef Page -> Bool -> IO (ConnectId Button)
eEludeToken bShrink bToken bookRef pageRef isFromL = Gtk.on bShrink buttonPressEvent $ do
  liftIO $ do
    book <- readIORef bookRef
    if (bkIsJapanese book) == True
    then
     do
      book <- readIORef bookRef
      page <- readIORef pageRef
      let
       isFromR = not isFromL
       cToken = bkCToken book
       cChars = takeFst $ pgChPs page :: V.Vector Char
       sideIndices@(mLeft, mRight) = bkCCharSideIndices book :: (Maybe Int, Maybe Int)
       neoTokenEtNeoMerginal@(nTok,nMerg)
         | mLeft == Nothing && mRight == Nothing = ("", (Nothing, Nothing))
         | mLeft == Just 0 && isFromL = ((cChars V.! 0) : cToken, (Nothing, mRight))
         | mRight == Just (lenOfChars - 1) && isFromR = (cToken P.++ [(cChars V.! (lenOfChars - 1))], (mLeft, Nothing))
         | isFromL = case mLeft of
             Nothing -> (cToken, (Nothing, mRight))
             Just i -> ((cChars V.! i) : cToken, (Just (i - 1), mRight))
         | otherwise = case mRight of
             Nothing -> (cToken, (Nothing, mRight))
             Just i -> (cToken P.++ [(cChars V.! i)], (mLeft, Just (i + 1)))
            where
               lenOfChars = V.length cChars
      liftIO $ modifyIORef' bookRef (\x -> x {bkCToken = nTok, bkCCharSideIndices = nMerg})
      buttonSetLabel bToken nTok
      return True
    else
      return True

eMoveOtherPage :: MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IORef Books -> Gtk.Button -> Direction -> IO (ConnectId Gtk.Button)
eMoveOtherPage window cpallete mpallete bookRef pageRef booksRef wButOther dir =
  Gtk.on wButOther buttonPressEvent $ do
    liftIO $ do
      book <- readIORef bookRef
      if dir == LeftD
        then
         eMovePagePrim window cpallete mpallete bookRef pageRef booksRef (-1)  (\n -> n < 0, bkMaxPage book)
        else
         eMovePagePrim window cpallete mpallete bookRef pageRef booksRef 1 (\n -> n > bkMaxPage book, 0)
      balancePage2 window cpallete mpallete bookRef pageRef booksRef
      return True

ePurgeCConfig window cpallete pallete bookRef pageRef booksRef butColor =
  Gtk.on butColor buttonPressEvent $ do
   liftIO $ do
    book <- readIORef bookRef
    page <- readIORef pageRef
    let
      cConfig2 = bkCConfig2 book
      dummyCConfig = V.singleton "poppy, Red"
    modifyIORef' bookRef (\x -> x {bkCConfig2 = dummyCConfig})
    balancePage2 window cpallete pallete bookRef pageRef booksRef
    widgetHideAll $ mpWindow pallete
    saveCConfig book
    return True

initMainWindow :: IORef Book -> IORef Page -> IO MainWindow
initMainWindow bookRef pageRef = do
    page <- readIORef pageRef
    let
      pixbuf = pgPixbuf page
    window <- windowNew
    windowMaximize window
    book <- readIORef bookRef
    Gtk.on window objectDestroy mainQuit
    vbox <- vBoxNew False 0
    hbox <- hBoxNew False 0

    wButPrevPrim <- buttonNewWithLabel "prev"
    boxPackStart hbox wButPrevPrim PackGrow 0
    wButVanillaPrim <- buttonNewWithLabel "toPlain"
    boxPackStart hbox wButVanillaPrim PackGrow 0
    wButNextPrim <- buttonNewWithLabel "next"
    boxPackStart hbox wButNextPrim PackGrow 0

    let
      winWidth = 960
      winHeight = 1417
      underbarHeight = 30
    Gtk.set window [windowDefaultWidth := winWidth, windowDefaultHeight := winHeight,
                containerChild := vbox, containerBorderWidth := 0]
    widgetSetSizeRequest hbox winWidth underbarHeight

    scroll <- scrolledWindowNew Nothing Nothing
    scrolledWindowSetPolicy scroll PolicyAutomatic PolicyAutomatic

    boxPackEnd vbox hbox PackNatural 0
    boxPackStart vbox scroll PackGrow 0
    image <- imageNewFromPixbuf pixbuf
    imageEvent <- eventBoxNew
    containerAdd imageEvent image
    scrolledWindowAddWithViewport scroll imageEvent
    widgetSetEvents scroll [ButtonPressMask]
    miscSetAlignment image 0 0
    miscSetPadding image 0 0
    let
      mainw = CMainW {
         wMain = window
       , wVbox = vbox
       , wHbox = hbox
       , wScroll = scroll
       , wImage = image
       , wHeight = winHeight
       , wWidth = winWidth
       , wUnderBarHeight = underbarHeight
       , wButNext = wButNextPrim
       , wButPrev = wButPrevPrim
       , wButVanilla = wButVanillaPrim
       }
    return mainw

initBooketPage :: Books -> IO (Book, Page)
initBooketPage books = do
    (pageFileNames, cConfig2, pSTDataPaths, pParsedPaths, pParsedPathsFull, currPage, docName, clink, chPPaths) <- initializeBook books
    let
      pagePaths = V.map (\x -> "./boxBooks/" P.++ docName P.++ "/imgs/" P.++ x) pageFileNames
      lastPage = (V.length pageFileNames - 1)
      cToken = "empty"
    pSTDataPrim <- iVecFromFile $ pSTDataPaths V.! currPage
    let
      pSTData = toPSTData pSTDataPrim
    chPDataPrim <- iVecFromFile $ chPPaths V.! currPage
    let
      chPData = V.map (\x -> read x :: (Char, V.Vector (V.Vector SquareD))) chPDataPrim
      chPData2 = V.map (\csq@(c,sq) -> (c, V.head $ V.head sq)) chPData
    pParsed <- iVecFromFile $ pParsedPaths V.! currPage
    pixbufVanilla <- pixbufNewFromFile (pagePaths V.! currPage)
    height <- pixbufGetHeight pixbufVanilla
    width <- pixbufGetWidth pixbufVanilla
    let
      isLangJP = True
      posColors = abstractPosColors pSTData cConfig2 isLangJP
      pPPath = "./boxBooks/" P.++ docName P.++ "/chP/" P.++ docName P.++ "_" P.++ (show $ currPage) P.++ ".txt"
    cpDataPrim <- iVecFromFile pPPath
    let
      charSquares = V.map f $ V.concatMap (\x -> V.concatMap id x) $ takeSnd $ V.map (\x -> read x :: (Char, V.Vector (V.Vector SquareD))) cpDataPrim
        where
          f sq = sqDtoSq (fromIntegral height) (fromIntegral width) sq
    cConfig2Ref <- newIORef cConfig2
    pixbuf <- coloringColorImageRefBuf posColors pixbufVanilla
    let
      book = CBook {
             bkCConfig2 = cConfig2
           , bkPagePaths = pagePaths
           , bkDocName = docName
           , bkPSTDataPaths  = pSTDataPaths
           , bkParsedPaths = pParsedPaths
           , bkParsedPathsFull = pParsedPathsFull
           , bkChPPaths = chPPaths
           , bkCToken = cToken
           , bkPositionTemp = (0.0, 0.0)
           , bkCurrPage = currPage
           , bkMaxPage = lastPage
           , bkIsJapanese = isLangJP
           , bkCCharSideIndices = (Nothing, Nothing)
           }
      page = CPage {
             pgPixbuf = pixbuf
           , pgPSTData = pSTData
           , pgParsed = pParsed
           , pgHeight = height
           , pgWidth = width
           , pgChPs = chPData2
           }
    return (book, page)

adjustBufToWindow :: Pixbuf -> MainWindow -> IORef Books -> IO Pixbuf
adjustBufToWindow pixbufVanilla window booksRef = do
      bufHeight <- pixbufGetHeight pixbufVanilla
      bufWidth <- pixbufGetWidth pixbufVanilla
      windowSize@(wWidth, wHeight) <- windowGetSize $ wMain window
      let
        resized@(resizedHeight, resizedWidth)
          | isHeightFull = (tarHeight, floor $ ratioInH * (fromIntegral bufWidth))
          | otherwise = (floor $ ratioInW * (fromIntegral bufHeight), tarWidth)
             where
               tarWidth = 960 - 5
               tarHeight = 1417 - 35
               ratioInH = (fromIntegral tarHeight) / (fromIntegral bufHeight)
               ratioInW = (fromIntegral tarWidth) / (fromIntegral bufWidth)
               isHeightFull = (fromIntegral tarHeight) < ratioInW * (fromIntegral bufHeight)
      neoPixbuf <- pixbufScaleSimple pixbufVanilla resizedWidth resizedHeight InterpBilinear
      return neoPixbuf

eRelease :: MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IORef Books -> IO ()
eRelease window cpallete mpallete bookRef pageRef booksRef = do
    signalId <- Gtk.on (wScroll window) buttonReleaseEvent $ do
      xy@(eX, eY) <- eventCoordinates
      liftIO $ do
        book <- readIORef bookRef
        page <- readIORef pageRef
        bufHeight <- pixbufGetHeight (pgPixbuf page)
        bufWidth <- pixbufGetWidth (pgPixbuf page)
        let
          currPage = bkCurrPage book :: Int
          cToken = bkCToken book
          oriHeight = pgHeight page
          oriWidth = pgWidth page
          ratio = (fromIntegral oriHeight) / (fromIntegral bufHeight)
        pSTDataPrim <- iVecFromFile $ (bkPSTDataPaths book) V.! (bkCurrPage book)
        let
          pSTData = toPSTData pSTDataPrim
          ratio = (fromIntegral oriWidth) / (fromIntegral bufWidth) :: Double
          pageWidth = pgWidth page
          xyInted@(yRow, xCol) = (truncate $ (snd xy) * ratio, truncate $ (fst xy) * ratio) :: (Int, Int)
          sqTokens = fromSTDataToSqTokens pSTData
          cToken = bkCToken book
          isCatchNothing = cToken == ""
        if isCatchNothing
               then do
                -- widgetShowAll $ mpWindow mpallete
                 widgetHideAll $ cpWindow cpallete
                 let
                   mpWinHeight = mpHeight mpallete
                   mpWinWidth = mpWidth mpallete
                   destX
                    | destPrim < 0 = 0
                    | otherwise = destPrim
                       where
                         destPrim = (floor eX) - (div mpWinWidth 2)
                   destY
                    | destPrim < 0 = 0
                    | otherwise = destPrim
                       where
                         destPrim = (floor eY) - mpWinHeight
                 --windowMove (mpWindow mpallete) destX destY
                 modifyIORef' bookRef (\bk -> bk {bkCToken = cToken})
                 return True
               else do
                 --widgetHideAll $ mpWindow mpallete
                 widgetShowAll $ cpWindow cpallete
                 let
                   cpWinHeight = cpHeight cpallete
                   cpWinWidth = cpWidth cpallete
                   destX
                    | destPrim < 0 = 0
                    | otherwise = destPrim
                       where
                         destPrim = (floor eX) - (div cpWinWidth 2)
                   destY
                    | destPrim < 0 = 0
                    | otherwise = destPrim
                       where
                         destPrim = (floor eY) - cpWinHeight
                 windowMove (cpWindow cpallete) destX destY
                 buttonSetLabel (cpButShowToken cpallete) cToken
                 modifyIORef' bookRef (\bk -> bk {bkCToken = cToken})
                 return True
    return ()

abstractPosColors ::
     V.Vector (V.Vector (String, V.Vector Square))
  -> V.Vector String
  -> Bool
  -> V.Vector (Square, MyColor)
abstractPosColors stData cConfig isLangJP = posColors
  where
    forgotten = V.concatMap id stData :: V.Vector (String, V.Vector Square)
    forgotten2
      | isLangJP = V.map (\x -> ((concatString "" $ takeFst x), (vNub $ V.concatMap id $ takeSnd x))) stData
      | otherwise = V.map (\x -> ((concatString " " $ takeFst x), (vNub $ V.concatMap id $ takeSnd x))) stData
        where
          concatString delim xs
             | xs == V.empty = ""
             | otherwise = V.foldl' g (V.head xs) (V.tail xs)
                where
                  g = (\y -> \x -> y P.++ delim P.++ x)
    sqTokens :: V.Vector (Square, String)
    sqTokens = fromSTDataToSqTokensFlattened stData
    cConfigV = vNub $ V.map (\x -> f x) cConfig :: V.Vector (String, MyColor)
       where
         f :: String -> (String, MyColor)
         f str = res
           where
             bul = delimitAtWO2 ',' id str :: [String]
             res = (tToken, tColor)
                where
                  tToken = replace "  " " " $ bul P.!! 0
                  tColor = stringToMyColor $ replace " " "" $ bul P.!! 1
         cNub :: Eq a => V.Vector (a, MyColor) -> V.Vector (a, MyColor)
         cNub lis = V.foldl' f V.empty lis
              where
                f y x@(str, col)
                 | V.elem str (takeFst y) = y
                 | otherwise = (V.cons x y)
    posColorsMapped = V.map (\x -> retrColondSqSimple sqTokens x) cConfigV :: V.Vector (V.Vector Square, MyColor)
    posColorsCatenated :: V.Vector (Square, MyColor)
    posColorsCatenated = (V.concatMap (\x@(sqss, colo) -> V.map (\y -> (y, colo)) sqss ) posColorsMapped)
    posColorsNubbed = V.foldl' h2 V.empty posColorsCatenated
       where
         h2 :: V.Vector (Square, MyColor) -> (Square, MyColor) -> V.Vector (Square, MyColor)
         h2 y x@(sq, col)
           | V.elem sq sqss = y
           | otherwise = y V.++ (V.singleton x)
           where
             sqss = takeFst y :: V.Vector Square
    posColors :: V.Vector (Square, MyColor)
    posColors = posColorsNubbed

eTouch :: MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IO ()
eTouch window cpallete mpallete bookRef pageRef = do
 signalId <- Gtk.on (wScroll window) buttonPressEvent $ do
      xy@(eX, eY) <- eventCoordinates
      -- xy@(eX, eY) = (830.0, 450.0)
      liftIO $ do
        book <- readIORef bookRef
        page <- readIORef pageRef
        bufWidth <- pixbufGetWidth (pgPixbuf page)
        bufHeight <- pixbufGetHeight (pgPixbuf page)
        let
          isJapanese = bkIsJapanese book
          currPage = bkCurrPage book :: Int
          oriHeight = pgHeight page
          ratio = (fromIntegral oriHeight) / (fromIntegral bufHeight)
        pSTDataPrim <- iVecFromFile $ (bkPSTDataPaths book) V.! (bkCurrPage book)
        let
          chPDataCurr = pgChPs page
          chPDataInted = V.map (\x@(c,sq) -> (c, f sq)) chPDataCurr
            where
             f ((r1,c1),(r2,c2)) =
               ((prodDtoInt bufHeight (ratio * r1), prodDtoInt bufWidth (ratio * c1)),
                (prodDtoInt bufHeight (ratio * r2), prodDtoInt bufWidth (ratio * c2)))
             prodDtoInt i r = floor $ r * (fromIntegral i)
          sqChars = V.map (\x@(c,sq) -> (sq,c)) chPDataInted :: V.Vector (Square, Char)
          sqCharsIndexed = V.imap f sqChars
            where
              f i (sq, c) = (sq, (i, c))
        let
          pSTData = toPSTData pSTDataPrim
          pSTDataFlattenedPartly :: V.Vector (String, V.Vector Square)
          pSTDataFlattenedPartly = V.concatMap id pSTData
          pSTDataFlattened = V.concatMap (\x -> V.concatMap (\y@(c,sqs) -> V.map (\z -> (c, z)) sqs) x) pSTData
          pageWidth = pgWidth page
          pageHeight = pgHeight page
        let
          xyInted@(yRow, xCol) = (truncate $ (snd xy) * ratio, truncate $ (fst xy) * ratio) :: (Int, Int)
          sqTokens = fromSTDataToSqTokens pSTData
          cTokenRawZipped = vlookupByAllMaybeSelfReturn xyInted sqTokens (flip isIncludesPoint)
          cCharRawZipped = vlookupByAllMaybeSelfReturn xyInted sqCharsIndexed (flip isIncludesPoint)
        let
          cTokenRaw = takeSnd <$> cTokenRawZipped
          cTokenPrim = case cTokenRaw of
            Just vtoks ->
              if vtoks == V.empty
              then ""
              else V.foldl' (\y -> \x -> y P.++ " " P.++ x) (V.head vtoks) (V.tail vtoks)
            Nothing -> ""
          cCharRaw = takeSnd <$> cCharRawZipped
          indCChar@(ind, cChar) = case cCharRaw of
            Just ivtoks ->
              if ivtoks == V.empty
              then (Nothing, "")
              else
                let
                  ivToksHead@(i,vtok) = V.head ivtoks
                in
                  (Just i, [vtok])
            Nothing -> (Nothing, "")
          cToken
           | isJapanese = cChar
           | otherwise = cTokenPrim
        let
          incSideIndices = case ind of
            Nothing -> (Nothing, Nothing)
            Just i ->
              let
                res
                 | i == 0 && i == V.length sqCharsIndexed = (Nothing, Nothing)
                 | i == 0 = (Nothing, Just (i + 1))
                 | i == V.length sqCharsIndexed - 1 = (Just (i - 1), Nothing)
                 | otherwise = (Just (i - 1), Just (i + 1))
              in
                res
        modifyIORef' bookRef (\bk -> bk {bkCToken = cToken, bkPositionTemp = xy, bkCCharSideIndices = incSideIndices})
        return True
 return ()


data Books = CBooks {
    bksList :: V.Vector String
  , bksCurrBookIndex :: Int
  , bksConfig :: V.Vector String
  , bksDHeight :: Int
  , bksDWidth :: Int
  }

data Book = CBook {
    bkCConfig2 :: V.Vector String
  , bkPagePaths :: V.Vector String
  , bkDocName :: String
  , bkPSTDataPaths :: V.Vector String
  , bkChPPaths :: V.Vector String
  , bkParsedPaths :: V.Vector String
  , bkParsedPathsFull :: V.Vector String
  , bkCToken :: String
  , bkPositionTemp :: (Double, Double)
  , bkCurrPage :: Int
  , bkMaxPage :: Int
  , bkIsJapanese :: Bool
  , bkCCharSideIndices :: (Maybe Int, Maybe Int)
  }

data PFlavor = Adhoc | Cyclic | Gramatica | Primitive | Semantical | SemanticalOne
  deriving (Show, Eq)

data Page = CPage {
    pgPixbuf :: Gtk.Pixbuf
  , pgPSTData :: V.Vector (V.Vector (String, V.Vector Square))
  , pgChPs :: V.Vector (Char, SquareD)
  , pgParsed :: V.Vector String
  , pgHeight :: Int
  , pgWidth :: Int
  }

data MainWindow = CMainW {
    wMain :: Gtk.Window
  , wVbox :: VBox
  , wHbox :: HBox
  , wImage :: Gtk.Image
  , wScroll :: ScrolledWindow
  , wWidth :: Int
  , wHeight :: Int
  , wUnderBarHeight :: Int
  , wButNext :: Button
  , wButPrev :: Button
  , wButVanilla :: Button
  }

data Direction = LeftD | RightD
  deriving (Show, Eq)

balancePage2 :: MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IORef Books -> IO ()
balancePage2 window cpallete mpallete bookRef pageRef booksRef = do
  book <- readIORef bookRef
  page <- readIORef pageRef
  let
    cConfig2 = bkCConfig2 book
    neoPage = bkCurrPage book
    fileName = (\files -> files V.! neoPage) $ bkPagePaths book
    neoPath = (bkPSTDataPaths book) V.! neoPage
  neoPSTDataPrim <- iVecFromFile neoPath
  neoParsed <- iVecFromFile ((bkParsedPaths book) V.! neoPage)
  let
    neoPSTData = toPSTData neoPSTDataPrim
    isLangJP = bkIsJapanese book
    posColors = abstractPosColors neoPSTData cConfig2 isLangJP
  pixbufVanilla <- pixbufNewFromFile fileName
  height <- pixbufGetHeight pixbufVanilla
  width <- pixbufGetWidth pixbufVanilla
  let
    docName = (bkDocName book)
    pPPath = "./boxBooks/" P.++ docName P.++ "/chP/" P.++ docName P.++ "_" P.++ (show $ neoPage) P.++ ".txt"
  cpDataPrim <- iVecFromFile pPPath
  let
    chPData = V.map (\x -> read x :: (Char, V.Vector (V.Vector SquareD))) cpDataPrim
    chPData2 = V.map (\csq@(c,sq) -> (c, V.head $ V.head sq)) chPData
    charTokSquares = V.map (\x@(c, sqss) -> (c, V.map f $ V.concatMap id sqss))
      $ V.map (\x -> read x :: (Char, V.Vector (V.Vector SquareD))) cpDataPrim
        where
          f sq = sqDtoSq (fromIntegral height) (fromIntegral width) sq
    charSquares = V.concatMap id $ takeSnd charTokSquares
  neoPixbufPrim <- coloringColorImageRefBuf posColors pixbufVanilla
  let
    isExpanded = True
  neoPixbuf <-
    if isExpanded == False
    then return neoPixbufPrim
    else
      do
       neoPixbuf2 <- adjustBufToWindow neoPixbufPrim window booksRef
       return neoPixbuf2
  imageSetFromPixbuf (wImage window) neoPixbuf

  modifyIORef' bookRef (\x -> x {bkCurrPage = neoPage})
  modifyIORef' pageRef (\x -> x {
                           pgPixbuf = neoPixbuf,
                           pgPSTData = neoPSTData,
                           pgChPs = chPData2,
                           pgParsed = neoParsed,
                           pgHeight = height,
                           pgWidth = width})
  let
    viewTitleForLang
     | bkIsJapanese book == True = "  Lang := " P.++ "Japanese"
     | otherwise = ("  Lang := " P.++ "English")
    windowTitle = docName P.++ "   Page " P.++ (show $ bkCurrPage book) P.++  viewTitleForLang
  windowSetTitle (wMain window) windowTitle

  saveCConfig book
  return ()


data ColorPallete = CColorPallete {
    cpWindow :: Gtk.Window
  , cpVBox :: VBox
  , cpButRed :: Button
  , cpButBlue :: Button
  , cpButGreen :: Button
  , cpButOrange :: Button
  , cpButPurple :: Button
  , cpButPink :: Button
  , cpButBrown :: Button
  , cpButAqua :: Button
  , cpButShowToken :: Button
  , cpButEludeToL :: Button
  , cpButEludeToR :: Button
  , cpHeight :: Int
  , cpWidth :: Int
  }

data MovePallete = CMovePallete {
    mpWindow :: Gtk.Window
  , mpVBox :: VBox
  , mpButHide :: Button
  , mpHeight :: Int
  , mpWidth :: Int
  , mpButFlush :: Button
  }

initColorPallete :: MainWindow -> IORef Book -> IORef Page -> IORef Books -> IO ColorPallete
initColorPallete window bookRef pageRef booksRef = do
    pWindowPrim <- windowNewPopup
    book <- readIORef bookRef
    windowSetTypeHint pWindowPrim WindowTypeHintPopupMenu
    vboxTot     <- vBoxNew False 0
    hboxTot     <- hBoxNew False 0
    hboxColor     <- hBoxNew False 0
    hboxColor2     <- hBoxNew False 0
    hboxToken     <- hBoxNew False 0
    hboxClose     <- hBoxNew False 0
    hboxElude <- hBoxNew False 0

    let
      cpHeightPrim = 200
      cpWidthPrim = 220
      tokenBoxHeight = 50

    widgetSetSizeRequest hboxToken cpWidthPrim tokenBoxHeight
    widgetSetSizeRequest hboxElude cpWidthPrim tokenBoxHeight

    tileRed <- imageNewFromFile "colortiles/red.png"
    tileBlue <- imageNewFromFile "colortiles/blue.png"
    tileGreen <- imageNewFromFile "colortiles/green.png"
    tilePurple <- imageNewFromFile "colortiles/purple.png"
    tileOrange <- imageNewFromFile "colortiles/orange.png"
    tilePink <- imageNewFromFile "colortiles/pink.png"
    tileBrown <- imageNewFromFile "colortiles/brown.png"
    tileAqua <- imageNewFromFile "colortiles/aqua.png"

    pButRedPrim <- buttonNew
    buttonSetImage pButRedPrim tileRed
    boxPackStart hboxColor pButRedPrim PackGrow 0

    pButBluePrim <- buttonNew
    buttonSetImage pButBluePrim tileBlue
    boxPackStart hboxColor pButBluePrim PackGrow 0

    pButPinkPrim <- buttonNew
    buttonSetImage pButPinkPrim tilePink
    boxPackStart hboxColor pButPinkPrim PackGrow 0

    pButAquaPrim <- buttonNew
    buttonSetImage pButAquaPrim tileAqua
    boxPackStart hboxColor pButAquaPrim PackGrow 0

    pButGreenPrim <- buttonNew
    buttonSetImage pButGreenPrim tileGreen
    boxPackStart hboxColor2 pButGreenPrim PackGrow 0

    pButOrangePrim <- buttonNew
    buttonSetImage pButOrangePrim tileOrange
    boxPackStart hboxColor2 pButOrangePrim PackGrow 0

    pButPurplePrim <- buttonNew
    buttonSetImage pButPurplePrim tilePurple
    boxPackStart hboxColor2 pButPurplePrim PackGrow 0

    pButBrownPrim <- buttonNew
    buttonSetImage pButBrownPrim tileBrown
    boxPackStart hboxColor2 pButBrownPrim PackGrow 0

    pButShowTokenPrim <- buttonNewWithLabel "empty"
    boxPackStart hboxToken pButShowTokenPrim PackGrow 0

    pButEludeToLPrim <- buttonNewWithLabel "<-"
    boxPackStart hboxElude pButEludeToLPrim PackGrow 0
    pButEludeToRPrim <- buttonNewWithLabel "->"
    boxPackEnd hboxElude pButEludeToRPrim PackGrow 0

--    boxPackStart vboxTot hboxClose PackGrow 0
    boxPackStart vboxTot hboxColor PackGrow 0
    boxPackStart vboxTot hboxColor2 PackGrow 0
    boxPackStart vboxTot hboxToken PackGrow 0
    boxPackStart vboxTot hboxElude PackGrow 0

    boxPackStart hboxTot vboxTot PackGrow 0

    Gtk.set pWindowPrim [windowDefaultWidth := cpWidthPrim, windowDefaultHeight := cpHeightPrim,
                  containerChild := hboxTot, containerBorderWidth := 0]

    let
     palleteW = CColorPallete {
       cpWindow = pWindowPrim
     , cpVBox = vboxTot
     , cpButRed = pButRedPrim
     , cpButGreen = pButGreenPrim
     , cpButBlue = pButBluePrim
     , cpButOrange = pButOrangePrim
     , cpButPurple = pButPurplePrim
     , cpButPink = pButPinkPrim
     , cpButBrown = pButBrownPrim
     , cpButAqua = pButAquaPrim
     , cpButShowToken = pButShowTokenPrim
     , cpHeight = cpHeightPrim
     , cpWidth = cpWidthPrim
     , cpButEludeToL = pButEludeToLPrim
     , cpButEludeToR = pButEludeToRPrim
     }
    return palleteW

initMovePallete :: MainWindow -> IORef Book -> IORef Page -> IORef Books -> IO MovePallete
initMovePallete window bookRef pageRef booksRef = do
    book <- readIORef bookRef
    page <- readIORef pageRef
    pWindowPrim <- windowNewPopup
    windowSetTypeHint pWindowPrim WindowTypeHintPopupMenu
    vboxTot     <- vBoxNew False 0

    let
      rightBoxWidth = 50
      rightBoxHeight = 30

    pButHidePrim <- buttonNewWithLabel "×"

    flushButton <- buttonNewWithLabel "Fl"

    books <- readIORef booksRef

    boxPackStart vboxTot pButHidePrim PackGrow 0
    boxPackStart vboxTot flushButton PackGrow 0

    let
      mpHeightPrim = 150
      mpWidthPrim = 150

    Gtk.set pWindowPrim [windowDefaultWidth := mpWidthPrim, windowDefaultHeight := mpHeightPrim,
                  containerChild := vboxTot, containerBorderWidth := 0]
    let
     palleteW = CMovePallete {
       mpWindow = pWindowPrim
     , mpVBox = vboxTot
     , mpButHide = pButHidePrim
     , mpHeight = mpHeightPrim
     , mpWidth = mpWidthPrim
     , mpButFlush = flushButton
     }
    return palleteW

sqDtoSq :: Double -> Double -> SquareD -> Square
sqDtoSq rr rc sqD@((r1, c1), (r2, c2)) = res
  where
    res = ((1 + (floor $ r1 * rr), 1 + (floor $ c1 * rc)), (1 + (floor $ r2 * rr), 1 + (floor $ c2 * rc)))

initializeBook books = do
  let
    docName = (bksList books) V.! (bksCurrBookIndex books)
    docRoot = "./boxBooks/" P.++ docName
  pagePaths <- getSpecExtFiles ("\"" P.++ docRoot P.++ "/imgs\"") "png"
  configFiles <- getSpecExtFiles ("\"" P.++ docRoot P.++ "/coConfigs\"") "txt"
  let
    pagePathsPrefix = V.map (\x -> P.take (P.length x - 4) x) pagePaths
    pagePathsPrefix2 = V.map (\x -> docName P.++ "_" P.++ (show (read (P.takeWhile (\c -> not $ c == '.' ) $ P.reverse $ P.takeWhile (\c -> not $ c == '_') $ P.reverse x) -1 :: Int))) pagePaths
    configFile
      | res == V.empty = ""
      | otherwise = V.head res
         where
           res = V.filter (\x -> x == docName P.++ "_cConfigRef.txt") configFiles
    retrConfigFile configFile
     | configFile == "" = f
     | otherwise = g
       where
         cConfigPath = "./boxBooks/" P.++ docName P.++ "/coConfigs/" P.++ docName P.++ "_cConfigRef.txt"
         f = do
           cConfigPrim <- iVecFromFile cConfigPath
           return cConfigPrim
         g = do
           cConfigPrim <- iVecFromFile cConfigPath
           return cConfigPrim
  cConfigPrim2 <- retrConfigFile configFile
  let
    cConfigHead = V.head cConfigPrim2
    cConfigTail = V.tail cConfigPrim2
    clink = read $ replace "\"" "" $ V.head cConfigTail :: V.Vector Int
    cConfig2 = V.map (\x -> read $! x) $ V.tail cConfigTail :: V.Vector String
    cCount2 = read $ replace "\"" "" $ cConfigHead :: Int
    pSTDataPaths = V.map (\x -> docRoot P.++ "/stData/" P.++ x P.++ ".txt") pagePathsPrefix
    pParsedPaths = V.map (\x -> docRoot P.++ "/stData/" P.++ x P.++ "_parsed.txt") pagePathsPrefix
    pParsedPathsFull = V.map (\x -> docRoot P.++ "/stData/" P.++ x P.++ "_withTag.txt") pagePathsPrefix
  (pSTDataPrim :: V.Vector (V.Vector String)) <- V.mapM (\x -> iVecFromFile $ docRoot P.++ "/stData/" P.++ x P.++ ".txt") pagePathsPrefix
  let
    chPPaths = V.map (\x -> docRoot P.++ "/chP/" P.++ x P.++ ".txt") pagePathsPrefix2
    pSTData = V.map f pSTDataPrim
        where
          f x = V.map (\y -> read $ y) x :: V.Vector (V.Vector ((String, V.Vector Square)))
    initCount
     | cCount2 < V.length pagePaths = cCount2
     | otherwise = 0
  return (pagePaths, cConfig2, pSTDataPaths, pParsedPaths, pParsedPathsFull, initCount, docName, clink, chPPaths)

toPSTData :: V.Vector String -> V.Vector (V.Vector ((String, V.Vector Square)))
toPSTData x = V.map (\y -> read $ y) x :: V.Vector (V.Vector ((String, V.Vector Square)))

initializeBooks = do
  dirs <- getLSRes "./boxBooks"
  bksConfigPrim <- iVecFromFile "bksConfig.txt"
  let
    bks = CBooks {
       bksList = dirs
     , bksCurrBookIndex = 0
     -- , bksCurrBookIndex = cIndex
     -- , bksConfig = bksConfigPrim
     , bksConfig = undefined
     , bksDWidth = 0
     , bksDHeight = 0
     }
  return bks

eMovePagePrim ::  MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IORef Books -> Int -> ((Int -> Bool), Int) -> IO ()
eMovePagePrim window cpallete pallete bookRef pageRef booksRef nOfSwipe (ifTerm, toPage) = do
  book <- readIORef bookRef
  let
    cConfig2 = bkCConfig2 book
    currPage = bkCurrPage book
    neoPage
     | ifTerm neoPrim = toPage
     | otherwise = neoPrim
        where
          neoPrim = currPage + nOfSwipe
  modifyIORef' bookRef (\bk -> bk {bkCurrPage = neoPage})
  balancePage2 window cpallete pallete bookRef pageRef booksRef
  saveCConfig book

eMovePage :: Gtk.Button -> MainWindow -> ColorPallete -> MovePallete -> IORef Book -> IORef Page -> IORef Books-> Int -> ((Int -> Bool), Int) -> IO (ConnectId Button)
eMovePage button window cpallete pallete bookRef pageRef booksRef nOfSwipe (ifTerm, toPage) = Gtk.on button buttonActivated $ do
  eMovePagePrim window cpallete pallete bookRef pageRef booksRef nOfSwipe (ifTerm, toPage)

saveCConfig :: Book -> IO ()
saveCConfig book = do
    let
      pagePaths = bkPagePaths book
      docName = bkDocName book
      tPageNum = bkCurrPage book
      tLinks = V.empty :: V.Vector Int
      tConfig = bkCConfig2 book
      sPath = "boxBooks/" P.++ docName P.++ "/coConfigs/" P.++ docName P.++ "_cConfigRef.txt"
      h2 x = P.dropWhile (\c -> c == '0') $ P.reverse $ P.takeWhile (\c -> not $ c == '_') $ P.reverse x
      fListsCaps = V.map h2 pagePaths
      headPage = V.head fListsCaps
      lastPage = V.last fListsCaps
      filePrefix = (P.takeWhile (\c -> not $ c == '_') $ V.head pagePaths) P.++ "_"
      cConfigOut = V.cons (show tPageNum) $ V.cons (show tLinks) tConfig
    oVecToFile cConfigOut sPath

saveCConfigBooks :: Books -> IO ()
saveCConfigBooks books = do
    let
      initBookID = bksCurrBookIndex books
      savStrs = V.singleton initBookID
      savPath = "bksConfig.txt"
    oVecToFile savStrs savPath

data CBCConfig = CCBCConfig {cbCB :: ComboBox, cbButDel :: Button}

eRegistCConfig window pallete mpallete bookRef pageRef booksRef butColor strColor =
  Gtk.on butColor buttonPressEvent $ do
   liftIO $ do
    book <- readIORef bookRef
    page <- readIORef pageRef
    let
      currPage = bkCurrPage book
      cToken = bkCToken book
      cConfig2 = bkCConfig2 book
    let
      updateCConfig :: String -> String -> V.Vector String -> V.Vector String
      updateCConfig col tok cConfig2
        | tok == "" = cConfig2
        | cConfig2 == V.empty = V.empty
        | otherwise = vNub $ V.snoc cConfig2 (tok P.++ ", " P.++ col)

    modifyIORef' bookRef (\x -> x {bkCConfig2 = (updateCConfig strColor cToken cConfig2)})
    balancePage2 window pallete mpallete bookRef pageRef booksRef
    widgetHideAll $ cpWindow pallete
    saveCConfig book
    return True

vMaximum :: Ord a => V.Vector a -> a
vMaximum xs = P.maximum $ V.toList xs

vAverage :: Integral a => V.Vector a -> Double
vAverage xs = (fromIntegral sumXs) / (fromIntegral lenXs)
  where
    lenXs = V.length xs
    sumXs = P.sum $ V.toList xs

vElemBy :: a -> V.Vector b -> (a -> b -> Bool) -> Bool
vElemBy x ys isEq = ors $ V.map (\z -> isEq x z) ys


takeNs :: Int -> V.Vector (V.Vector a) -> V.Vector a
takeNs n xss = V.map (\xs -> xs V.! n) xss


mergeVector :: Ord a => Ord b => V.Vector (a, V.Vector b) -> V.Vector (a, V.Vector b)
mergeVector vs = res
  where
    sorted = vSortBy (\x@(x1,_) -> \y@(y1,_) -> compare x1 y1) vs  -- :: V.Vector (a, Int)
    res
     | sorted == V.empty = V.empty
     | otherwise = poled V.++ (V.singleton st)
       where
         res2@(st, poled) = V.foldl' f (V.head sorted, V.empty) $ V.tail sorted
           where
             f y@(stacked@(chsymY, chsY), pooled) x@(chsymX, chsX)
              | chsymY == chsymX = ((chsymX, chsY V.++ chsX), pooled)
              | otherwise = (x, pooled V.++ (V.singleton stacked))

p2DtoPRawImg :: Int -> Int -> (Int, Int) -> Int
p2DtoPRawImg row col point@(pRow, pCol) = pRow * col + pCol

pSqtoPRawImg :: Int -> Int -> Square -> Vector Int
pSqtoPRawImg row col sq@((r1,c1),(r2,c2)) = res
  where
    sqPoints = sqToPSq row col sq
    res = map (\x -> p2DtoPRawImg row col x) sqPoints

sqToPSq :: Int -> Int -> Square -> Vector (Int, Int)
sqToPSq row col sq@((r1,c1),(r2,c2)) = sqPoints
  where
    q = col
    sqPoints = fromList [(r, c) | r <- [r1 .. r2], c <- [c1 .. c2]]



boolToColor :: (Vector Bool, (Int, Int)) -> (Vector (Pixel8, Pixel8, Pixel8), (Int, Int))
boolToColor = liftBoolImage

colorToBool :: (Vector (Pixel8, Pixel8, Pixel8), (Int, Int)) -> (Vector Bool, (Int, Int))
colorToBool = imageBoolize

coloringColorImageRefBuf :: V.Vector (Square, MyColor) -> Pixbuf -> IO Pixbuf
coloringColorImageRefBuf posColors pixbuf = updateBuf pixbuf table2
  where
    charsPosGWithPixels :: Vector (Square, (Pixel8, Pixel8, Pixel8))
    charsPosGWithPixels =
      fromBoxToUnbox $
      V.map (\x@(x1, x2) -> (x1, fromColor x2)) posColors
    table2 :: Vector ((Int, Int), (Pixel8, Pixel8, Pixel8))
    table2 = Data.Vector.Unboxed.concatMap f charsPosGWithPixels
      where
        f x@(sq@((r1, c1), (r2, c2)), color) = res
          where
            rows = enumFromN r1 (r2 - r1 + 1) :: Vector Int
            cols = enumFromN c1 (c2 - c1 + 1) :: Vector Int
            points =
              Data.Vector.Unboxed.concatMap
                (\n -> map (\r -> (n, r)) cols)
                rows :: Vector (Int, Int)
            res = map (\x -> (x, color)) points

updateBuf :: Pixbuf -> Vector ((Int, Int), (Pixel8, Pixel8, Pixel8)) -> IO Pixbuf
updateBuf pixbuf table2 = do
  (pixbufPixels :: PixbufData Int Pixel8) <- pixbufGetPixels pixbuf
  rowStrideBytes <- pixbufGetRowstride pixbuf
  nChannels <- pixbufGetNChannels pixbuf
  hasAlpha <- pixbufGetHasAlpha pixbuf
  bitsPerSample <- pixbufGetBitsPerSample pixbuf
  let
    --nChannels = 3
    rowStride = rowStrideBytes `quot` nChannels
    coloringBuf :: ((Int, Int), (Pixel8, Pixel8, Pixel8)) -> IO ()
    coloringBuf ((i, j), (r, g, b)) = do
      let
        p = i * rowStrideBytes + (j * nChannels)
      cR <- MArr.readArray pixbufPixels (p + 0)
      cG <- MArr.readArray pixbufPixels (p + 1)
      cB <- MArr.readArray pixbufPixels (p + 2)
      if cR < 255 && cG < 255 && cB < 255
        then
         do
          MArr.writeArray pixbufPixels (p + 0) $ r
          MArr.writeArray pixbufPixels (p + 1) $ g
          MArr.writeArray pixbufPixels (p + 2) $ b
        else
         return ()
  Data.Vector.Unboxed.mapM_ coloringBuf table2
  return pixbuf

main :: IO ()
main = do
  mainGtk

iVecFromFile :: String -> IO (V.Vector String)
iVecFromFile listPath = do
  handle <- openFile listPath ReadMode
  fText <- hGetContents handle
  let forout = V.fromList $ lines fText :: V.Vector String
  return forout

oVecToFile :: Show a => V.Vector a -> String -> IO ()
oVecToFile vec fpath = do
  handle <- openFile fpath WriteMode
  V.mapM (\x -> hPutStrLn handle $ show x) vec
  hClose handle

vDiff :: Eq a => V.Vector a -> V.Vector a -> V.Vector a
vDiff as1 as2 = V.filter (\x -> not (V.elem x as2)) as1

vNub :: Eq a => V.Vector a -> V.Vector a
vNub xs = V.fromList $ Lis.nub $ V.toList xs

vNubBy :: Eq a => (a -> a -> Bool) -> V.Vector a -> V.Vector a
vNubBy f xs = V.fromList $ Lis.nubBy f $ V.toList xs

ors :: V.Vector Bool -> Bool
ors bs = V.foldr (||) False bs

ands :: V.Vector Bool -> Bool
ands bs = V.foldr (&&) True bs

isOrs :: Eq a => a -> V.Vector a -> Bool
isOrs x xs = ors $ V.map (== x) xs

isAnds :: Eq a => a -> V.Vector a -> Bool
isAnds x xs = ands $ V.map (== x) xs


deMaybe :: a -> Maybe a -> a
deMaybe defo val =
  case val of
    Nothing -> defo
    Just v -> v


delimitAtWO2 :: Eq a => a -> ([a] -> b) -> [a] -> [b]
delimitAtWO2 del conv bulks =
  (conv $ snd res) : (fst res)
   where
     res = P.foldr (f del conv) ([],[]) bulks
       where
         f :: Eq a => a -> ([a] -> b)  -> a -> ([b], [a]) -> ([b], [a])
         f d cv a ([], []) = if a == d then ([], []) else ([], [a])
         f d cv a (bs, as) = if a == d
           then ((cv as) : bs, []) else (bs, a : as)

vlookupByAllMaybeSelfReturn :: Eq a => Eq b => Eq c => a -> V.Vector (c, b) -> (a -> c -> Bool) -> Maybe (V.Vector (c, b))
vlookupByAllMaybeSelfReturn query list f =
  let
    aa = V.foldl' (g query) V.empty list
       where
         g que y x@(xa, xb)
          | (f que xa) == True = V.snoc y x
          | otherwise = y
  in
   if
     aa == V.empty
   then Nothing
   else (Just aa)

cshow :: Show a => V.Vector a -> IO (V.Vector ())
cshow vec = do
  V.mapM (\x -> putStrLn $ show x) vec

cshowi :: Show a => V.Vector a -> IO (V.Vector ())
cshowi vec = do
  V.mapM (\x@(x1, x2) -> putStrLn $ (show x1) P.++ "     " P.++ (show x2)) $ indexing vec

cshowI :: Show a => V.Vector a -> IO (V.Vector ())
cshowI vs = cshow $ indexing vs

indexing :: V.Vector a -> V.Vector (Int, a)
indexing xs = V.zip (V.fromList [0 .. (V.length xs - 1)]) xs

vSortBy :: Ord a => (a -> a -> Ordering) -> V.Vector a -> V.Vector a
vSortBy f xs = V.fromList $  Lis.sortBy f $ V.toList xs

fromSTDataToSqTokens :: V.Vector (V.Vector (String, V.Vector Square))
  -> V.Vector (Square, String)
fromSTDataToSqTokens stData = V.concatMap f stData
  where
    f :: V.Vector (String, V.Vector Square) -> V.Vector (Square, String)
    f stDatum = V.concatMap (\x@(str, sqs) -> V.map (\y -> (y, str)) sqs) stDatum

isIncludesPoint :: Square -> Main.Point -> Bool
isIncludesPoint sq@((r3, c3), (r4, c4)) term@(r1, c1) = r3 < r1 && r1 < r4 && c3 < c1 && c1 < c4

fromSTDataToSqTokensFlattened :: V.Vector (V.Vector (String, V.Vector Square))
  -> V.Vector (Square, String)
fromSTDataToSqTokensFlattened stData = res
  where
    concated = V.map f stData
      where
        f stDatum = (sqs, token)
          where
            token = fromTokensToSens $ takeFst stDatum
            sqs = V.concatMap id $ takeSnd stDatum
    res = V.concatMap (\x@(sqs, tok) -> V.map (\y -> (y, tok)) sqs) concated

retrColondSqSimple :: V.Vector (Square, String) -> (String, MyColor)
  -> (V.Vector Square, MyColor)
retrColondSqSimple sqTokens cConfigV@(tokn, color) = (res, color)
  where
    res = takeFst $ V.filter (\x@(_, str) -> f tokn str) sqTokens
      where
        f tokn str = Lis.isInfixOf (P.map toLower tokn) (P.map toLower str)

getSpecExtFiles :: String -> String -> IO (V.Vector String)
getSpecExtFiles pngPath ext = do
  files <- getLSRes pngPath
  let
    pngFilesFiltered = V.filter f files
      where
        f xs = isExtOf ext xs
  return pngFilesFiltered

isExtOf :: String -> String -> Bool
isExtOf ext fileName
 | P.length ext == P.length fileName = False
 | otherwise = ext == fileExt
   where
     fileExt = P.reverse $ P.takeWhile (\c -> not $ c == '.') $ P.reverse fileName

data MyColor
  = Red
  | Green
  | Blue
  | White
  | Black
  | Gray
  | Navy
  | Lime
  | Aqua
  | Yellow
  | Pink
  | Olive
  | Purple
  | Brown
  | Orange
  | Gray2
  | LRed
  | LGreen
  | LBlue
  | LWhite
  | LBlack
  | LGray
  | LNavy
  | LLime
  | LAqua
  | LYellow
  | LPink
  | LOlive
  | LPurple
  | LBrown
  | LOrange
  | LGray2
  | Dum
  deriving (Show, Read, Eq, Ord)

fromColor :: MyColor -> (Pixel8, Pixel8, Pixel8)
fromColor Red = (255, 0, 0)
fromColor Green = (0, 128, 0)
fromColor Blue = (0, 0, 255)
fromColor White = (255, 255, 255)
fromColor Black = (0, 0, 0)
fromColor Gray = (128, 128, 128)
fromColor Navy = (0, 0, 128)
fromColor Lime = (0, 255, 0)
--fromColor Aqua = (0, 255, 255)
fromColor Aqua = (0, 200, 237)
fromColor Yellow = (255, 255, 0)
fromColor Pink = (255, 0, 255)
fromColor Olive = (128, 128, 0)
fromColor Purple =(128, 0, 128)
fromColor Brown = (128, 0, 0)
fromColor Orange = (255, 165, 0)
--fromColor Orange = (255, 130, 0)
fromColor Gray2 = (51, 51, 51)
fromColor LRed = mapTupT (\x -> div x 2) (255, 0, 0)
fromColor LGreen = mapTupT (\x -> div x 2) (0, 128, 0)
fromColor LBlue = mapTupT (\x -> div x 2) (0, 0, 255)
fromColor LWhite = mapTupT (\x -> div x 2) (255, 255, 255)
fromColor LBlack = mapTupT (\x -> div x 2) (0, 0, 0)
fromColor LGray = mapTupT (\x -> div x 2) (128, 128, 128)
fromColor LNavy = mapTupT (\x -> div x 2) (0, 0, 128)
fromColor LLime = mapTupT (\x -> div x 2) (0, 255, 0)
fromColor LAqua = mapTupT (\x -> div x 2) (0, 255, 255)
fromColor LYellow = mapTupT (\x -> div x 2) (255, 255, 0)
fromColor LPink = mapTupT (\x -> div x 2) (255, 0, 255)
fromColor LOlive = mapTupT (\x -> div x 2) (128, 128, 0)
fromColor LPurple = mapTupT (\x -> div x 2) (128, 0, 128)
fromColor LBrown = mapTupT (\x -> div x 2) (128, 0, 0)
fromColor LOrange = mapTupT (\x -> div x 2) (255, 165, 0)
fromColor LGray2 = mapTupT (\x -> div x 2) (51, 51, 51)
fromColor Dum = (0, 0, 0)


mapTupT :: (a -> b) -> (a, a, a) -> (b, b, b)
mapTupT f tup@(t1, t2, t3) = (f t1, f t2, f t3)

stringToMyColor :: String -> MyColor
stringToMyColor "Red" = Red
stringToMyColor "Green" = Green
stringToMyColor "Blue" = Blue
stringToMyColor "White" = White
stringToMyColor "Black" = Black
stringToMyColor "Gray" = Gray
stringToMyColor "Navy" = Navy
stringToMyColor "Lime" = Lime
stringToMyColor "Aqua" = Aqua
stringToMyColor "Yellow" = Yellow
stringToMyColor "Pink" = Pink
stringToMyColor "Olive" = Olive
stringToMyColor "Purple" = Purple
stringToMyColor "Brown" = Brown
stringToMyColor "Orange" = Orange
stringToMyColor "Gray2" = Gray2
stringToMyColor "LRed" = LRed
stringToMyColor "LGreen" = LGreen
stringToMyColor "LBlue" = LBlue
stringToMyColor "LWhite" = LWhite
stringToMyColor "LBlack" = LBlack
stringToMyColor "LGray" = LGray
stringToMyColor "LNavy" = LNavy
stringToMyColor "LLime" = LLime
stringToMyColor "LAqua" = LAqua
stringToMyColor "LYellow" = LYellow
stringToMyColor "LPink" = LPink
stringToMyColor "LOlive" = LOlive
stringToMyColor "LPurple" = LPurple
stringToMyColor "LBrown" = LBrown
stringToMyColor "LOrange" = LOrange
stringToMyColor "LGray2" = LGray2
stringToMyColor "Dum" = Dum

takeFst :: V.Vector (a, b) -> V.Vector a
takeFst ab = V.map (\x -> (fst x)) ab

takeSnd :: V.Vector (a, b) -> V.Vector b
takeSnd ab = V.map (\x -> (snd x)) ab

takeFstT :: V.Vector (a, b, c) -> V.Vector a
takeFstT abc = V.map f abc
    where
      f (a, b, c) = a

takeSndT :: V.Vector (a, b, c) -> V.Vector b
takeSndT abc = V.map f abc
    where
      f (a, b, c) = b

takeThdT :: V.Vector (a, b, c) -> V.Vector c
takeThdT abc = V.map f abc
    where
      f (a, b, c) = c

fstT :: (a, b, c) -> a
fstT (x, y, z) = x

sndT :: (a, b, c) -> b
sndT (x, y, z) = y

thdT :: (a, b, c) -> c
thdT (x, y, z) = z

fromUnboxToBox :: Unbox a => Vector a -> V.Vector a
fromUnboxToBox rawimg = V.fromList $ toList rawimg

fromBoxToUnbox :: Unbox a => V.Vector a -> Vector a
fromBoxToUnbox rawimg = fromList $ V.toList rawimg

liftBool :: Bool -> (Pixel8, Pixel8, Pixel8)
liftBool n = (val, val, val)
   where
     val = fromIntegral (boolToMono n)

liftBoolImage ::
  (Vector Bool, (Int, Int))
  -> (Vector (Pixel8, Pixel8, Pixel8), (Int ,Int))
liftBoolImage (monoimg, rowcol) =
  (map liftBool monoimg, rowcol)

boolToMono :: Bool -> Int
boolToMono val = if val == True then 255 else 0

imageBoolize ::
  (Vector (Pixel8, Pixel8, Pixel8), (Int ,Int))
  -> (Vector Bool, (Int, Int))
imageBoolize (rawimg, rowcol) =
  (map boolize rawimg, rowcol)

boolize :: (Pixel8, Pixel8, Pixel8) -> Bool
boolize (r, g, b) =
  if (fromIntegral r + fromIntegral g + fromIntegral b) `div` 3 < 200
  then False
  else True

type SquareD = ((Double, Double), (Double, Double))

type MonoImage = (Vector Bool, (Int, Int))
type BlackPoints = V.Vector (Int, Int)
type Square = ((Int, Int), (Int, Int))
type Squares = V.Vector ((Int, Int), (Int, Int))
type Row = Int
type Col = Int
type Sentense = V.Vector String
type ColorImage = (Vector (Pixel8, Pixel8, Pixel8), (Int, Int))

getLSRes :: String -> IO (V.Vector String)
getLSRes path = getShellRes $ "ls " P.++ path

getShellRes :: String -> IO (V.Vector String)
getShellRes cmd = do
  let
    texized = Text.pack cmd
  res2 <- fmap (Text.unpack . lineToText) <$> get2 texized
  let
    res3 = V.fromList res2
  return res3

get2 :: (MonadIO m) => Text -> m [Line]
get2 cmd = do
  fold (inshell cmd Turtle.empty) Fold.list

fromTokensToSens :: V.Vector String -> String
fromTokensToSens arr
  | arr == V.empty = ""
  | otherwise = V.foldl' f (V.head arr) (V.tail arr)
    where
      f y x = y P.++ " " P.++ x

iVecFromFileJP :: String -> IO (V.Vector String)
iVecFromFileJP listPath = do
  handle <- openFile listPath ReadMode
  hSetEncoding handle utf8
  fText <- hGetContents handle
  let forout = V.fromList $ lines fText :: V.Vector String
  return forout

oVecToFileJP :: Show a => V.Vector a -> String -> IO ()
oVecToFileJP vec fpath = do
  handle <- openFile fpath WriteMode
  hSetEncoding handle utf8
  V.mapM (\x -> hPutStrLn handle $ ushow x) vec
  hClose handle

oStrToFileJP :: String -> String -> IO ()
oStrToFileJP str fpath = do
  handle <- openFile fpath WriteMode
  hSetEncoding handle utf8
  hPutStr handle str
  hClose handle

cshowR :: Show a => V.Vector a -> IO (V.Vector ())
cshowR vs = cshow $ V.reverse $ indexing vs

cshowS :: Show a => V.Vector a -> Int -> Int-> IO (V.Vector ())
cshowS vs n m = cshow forShow
  where
    indexed = indexing vs
    forShow = V.map (\n -> indexed V.! n) $ V.fromList [n .. m]

cshowU :: Show a => V.Vector a -> IO (V.Vector ())
cshowU vec = do
  V.mapM (\x -> putStrLn $ ushow x) vec

cshowUI :: Show a => V.Vector a -> IO (V.Vector ())
cshowUI vs = cshowU $ indexing vs
